/*
 * A pipe paging coordinator, permitting memory sharing pipe endpoints.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/irq.h>

#include <ipc/cap_alloc.h>
#include <ipc/irq.h>
#include <mem/memory_preallocated.h>

#include "page_queue_partitioned.h"
#include "pipe_paging.h"



PipePaging::PipePaging(Memory *memory, offset_t size)
: _memory(NULL), _size(size)
{
    /* Reserve space for two pipe regions. */

    _memory = new MemoryPreallocated(memory, round(size, memory->region_size()) * 2);
    _queue = new PageQueuePartitioned();
    _pages = new Pages(_memory, _queue);

    /* Reset the mappers associated with the regions. */

    for (unsigned int i = 0; i < 2; i++)
        _regions[i] = NULL;

    /* Initialise IRQ objects and flags for notifications. */

    for (unsigned int i = 0; i < 2; i++)
    {
        _irqs[i] = L4_INVALID_CAP;
        _flags[i] = 0;
    }
}

/* Create an IRQ to subscribe to an endpoint's notifications. */

l4_cap_idx_t PipePaging::subscribe(bool writing, notify_flags_t flags)
{
    int i = writing ? 1 : 0;

    if (l4_is_invalid_cap(_irqs[i]))
        ipc_create_irq(&_irqs[i]);

    _flags[i] = flags;

    return _irqs[i];
}

/* Release any IRQ used for an endpoint's notifications. */

void PipePaging::unsubscribe(bool writing)
{
    int i = writing ? 1 : 0;

    if (l4_is_valid_cap(_irqs[i]))
    {
        ipc_cap_free_um(_irqs[i]);
        _irqs[i] = L4_INVALID_CAP;
        _flags[i] = 0;
    }
}

/* Notify the other endpoint. */

void PipePaging::notify(bool writing, notify_flags_t flags)
{
    /* Let the writer notify the reader, and the other way round. */

    int i = writing ? 0 : 1;

    if (l4_is_valid_cap(_irqs[i]) && (flags & _flags[i]))
        l4_irq_trigger(_irqs[i]);
}

/* Return whether one or more endpoints have detached. */

int PipePaging::closed()
{
    return _endpoints < 2;
}

void PipePaging::discard_region(unsigned int i)
{
    PageMapper *mapper = _regions[i];

    if (mapper != NULL)
    {
        mapper->detach();
        _regions[i] = NULL;
        delete mapper;
    }
}

/* Detach one endpoint. */

void PipePaging::detach()
{
    if (!_endpoints)
        return;
    else
        _endpoints--;

    /* Return if the other endpoint is attached. */

    if (_endpoints)
        return;

    /* Discard all regions from the pipe. */

    for (unsigned int i = 0; i < 2; i++)
        discard_region(i);

    /* Release IRQs. */

    for (unsigned int i = 0; i < 2; i++)
    {
        if (l4_is_valid_cap(_irqs[i]))
        {
            ipc_cap_free_um(_irqs[i]);
            _irqs[i] = L4_INVALID_CAP;
        }
    }

    /* Delete the page collection and related objects. */

    delete _pages;
    delete _queue;
    delete _memory;
}

/* Add a region to the sequence. */

PageMapper *PipePaging::add_region()
{
    /* If the writer already accesses a different region to the reader, no new
       region is added. */

    if (_writing != _reading)
        return NULL;

    /* Select the other region of the pair being maintained. */

    _writing = 1 - _writing;

    /* Make a new mapper for the region. */

    PageMapper *mapper = new PageMapper(&_accessors[_writing], _pages);

    /* Initialise and record the mapper. */

    mapper->attach();
    mapper->set_data_size(0);

    _regions[_writing] = mapper;

    /* Let the writer notify the reader. */

    notify(true, NOTIFY_CONTENT_AVAILABLE);

    /* Return the next region's mapper. */

    return mapper;
}

/* Return the current region for reading. */

PageMapper *PipePaging::current_region()
{
    return _regions[_reading];
}

/* Return the next region for the reader if the writer is using a different one.
   Otherwise, return NULL. */

PageMapper *PipePaging::next_region()
{
    /* If the reader already accesses the same region to the writer, no next
       region can be obtained. */

    if (_reading == _writing)
        return NULL;

    /* Detach and discard the current page mapper. */

    discard_region(_reading);

    /* Select the next region. */

    _reading = 1 - _reading;

    /* Let the reader notify the writer. */

    notify(false, NOTIFY_SPACE_AVAILABLE);

    /* Return the next region's mapper. */

    return _regions[_reading];
}

// vim: tabstop=4 expandtab shiftwidth=4
